"""
Module contains Python representations of Power BI objects.

The powerbi.PowerBI() client will attempt to translate API responses into instances 
of these objects.
"""

from dataclasses import dataclass, field
from typing import Optional

from dateutil.parser import parse

from .utils import camel_case_dict_keys


@dataclass(kw_only=True)
class PBIObject:
    raw: dict = field(repr=False, default=None)

    @classmethod
    def from_raw(cls, raw):
        """
        Create an instance of a `PBIObject` from the raw json response
        of an API Endpoint.

        Parameters
        ----------
        `raw` : `dict`
            Raw json response from the API call.

        Returns
        -------
        `PBIObject`
            Instance of a `PBIObject` type. The type returned depends on
            where it was called from.
        """

        kwargs = camel_case_dict_keys(raw)
        return cls(**kwargs, raw=raw)


@dataclass
class Group(PBIObject):
    """A Power BI group. Commonly called a Workspace."""

    id: str
    
    name: str = None
    type: str = None
    is_read_only: bool = None
    is_on_dedicated_capacity: bool = None
    capacity_id: str = None
    dataflow_storage_id: str = None


@dataclass
class Dataset(PBIObject):
    """A Power BI Dataset."""

    id: str
    
    name: str = None
    web_url: str = None
    add_rows_api_enabled: bool = None
    is_refreshable: bool = None
    is_effective_identity_required: bool = None
    is_effective_identity_roles_required: bool = None
    is_in_place_sharing_enabled: bool = None
    is_on_prem_gateway_required: bool = None
    target_storage_mode: str = None
    created_date: str = None
    create_report_embed_url: str = None
    qna_embed_url: str = None
    upstream_datasets: str = None
    users: Optional[list] = field(default_factory=list)
    configured_by: str = None

    def __post_init__(self):
        if self.created_date:
            self.created_date = parse(self.created_date)


@dataclass
class Refresh(PBIObject):
    """A Power BI refresh history entry."""
    
    id: int = None
    
    request_id: str = None
    refresh_type: str = None
    start_time: str = None
    end_time: str = None
    status: str = None
    service_exception_json: str = None

    def __post_init__(self):
        if self.start_time:
            self.start_time = parse(self.start_time)
        if self.end_time:
            self.end_time = parse(self.end_time)


# TODO: Consider creating subclassess for each subtype.
@dataclass
class ActivityEvent(PBIObject):
    """Power BI Auditing Events.
    
    Activity Events are audit and tracked activity on the Power BI 
    instance. Examples of Activity Events can include: viewing reports, 
    refreshing datasets, updating apps, etc.

    While there are many types of Activity Events, pybi consolidates these 
    into an `ActivityEvent` object. Attributes not related to an activity 
    type will be set to `None`.
    """
    
    id: str
    
    activity: str = None
    activity_id: str = None
    app_dashboard_id: str = None
    app_id: str = None
    app_name: str = None
    app_report_id: str = None
    artifact_id: str = None
    artifact_kind: str = None
    artifact_name: str = None
    client_ip: str = None
    consumption_method: str = None
    creation_time: str = None
    dashboard_id: str = None
    dashboard_name: str = None
    data_connectivity_mode: str = None
    dataflow_access_token_request_parameters: str = None
    dataflow_id: str = None
    dataflow_name: str = None
    dataflow_type: str = None
    dataset_id: str = None
    dataset_name: str = None
    datasets: str = None
    datasource_object_ids: str = None
    distribution_method: str = None
    exported_artifact_info: str = None
    folder_access_requests: str = None
    folder_display_name: str = None
    folder_object_id: str = None
    gateway_id: str = None
    has_full_report_attachment: bool = None
    import_display_name: str = None
    import_id: str = None
    import_source: str = None
    import_type: str = None
    is_success: bool = None
    item_name: str = None
    last_refresh_time: str = None
    model_id: str = None
    models_snapshots: str = None
    monikers: str = None
    object_id: str = None
    operation: str = None
    org_app_permission: str = None
    organization_id: str = None
    record_type: str = None
    refresh_type: str = None
    report_id: str = None
    report_name: str = None
    report_type: int = None
    request_id: str = None
    schedules: str = None
    share_link_id: str = None
    sharing_action: str = None
    sharing_information: str = None
    user_agent: str = None
    user_id: str = None
    user_key: str = None
    user_type: int = None
    work_space_name: str = None
    workload: str = None
    workspace_id: str = None

    def __post_init__(self):
        if self.creation_time:
            self.creation_time = parse(self.creation_time)
        if self.last_refresh_time:
            self.last_refresh_time = parse(self.last_refresh_time)



"""Utility functions used internally by pybi."""

import copy
import json
import re


def camel_to_snake(s):
    pattern = re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")
    return pattern.sub(r"_\1", s).lower()


def camel_case_dict_keys(d):
    return {camel_to_snake(k): v for k, v in d.items()}


def _convert_js_strings(js):
    for k, v in js.items():
        if isinstance(v, dict):
            _convert_js_strings(v)
        elif isinstance(v, str):
            try:
                js_from_str = json.loads(v)
                js[k] = js_from_str
            except:
                pass


def convert_js_strings(js):
    """
    Recursively traverse a dict and attempt to convert any json-like strings to json.

    Parameters
    ----------
    `js` : `dict`
        Json or dict to traverse and convert

    Returns
    -------
    `dict`
        New dict with json-like strings converted to json.
    """    

    new_js = copy.deepcopy(js)
    _convert_js_strings(new_js)

    return new_js



"""
Module implements a wrapper around the Power BI Rest API.

Users construct an instance of the PowerBI client and call its methods. These 
methods closely follow the naiming laid out here:

https://learn.microsoft.com/en-us/rest/api/power-bi/
"""

import requests

from .models import Group, Dataset, Refresh


class PowerBI:
    """
    User Interface into Power BI Rest Api.

    Users interact with their Power BI service by constructing an instance of this
    object and calling its methods.

    Authentication with the Power BI service requires a `bearer_token` which must be
    generated in advance before creating the `PowerBI` object. How you generate the token
    will depend on your Azure and Power BI configuration.

    In general, `PowerBI()` methods follow the naming laid out here:

    https://learn.microsoft.com/en-us/rest/api/power-bi/

    Parameters
    ----------
    `bearer_token` : `str`
        Bearer token used to authenticate with your Power BI service.
    """

    def __init__(self, bearer_token, session=None):
        self.bearer_token = bearer_token

        if not session:
            self.session = requests.Session()

        self.session.headers.update({"Authorization": f"Bearer {self.bearer_token}"})

    def _get_resource(self, url_format, *id, parameters={}):
        """
        Fetch a resource from the Power BI Rest API.

        Takes a templated url, id(s) used in the resource, and optional
        parameters. Constructs the url to the resource and requests the
        resource from the Power BI Rest API. Returns the associated response.

        If the response contains a list of Power BI objects, this will be
        parsed and the list of objects returned.

        Parameters
        ----------
        `url_format` : `str`
            The url of the resource. May also include placeholders,
            e.g., `datasets/{datasetId}/refreshes`.
        `*id` : `tuple`
            Id values to include in the url. Any placeholders in
            `url_format` will be replaced with these values in the
            order they are passed in.
        `**parameters`: `dict`
            Keyword argument of optional url parameters, e.g.,
            `$top` to limit number of results returned.


        Returns
        -------
        `json`
            The json response as returned from the api call. Where the api response contains
            a list of objects, e.g., a list of Datasets, the response will be parsed to only
            return this list.
        """

        if id:
            if isinstance(id, tuple):
                url = url_format.format(*id)
            else:
                url = url_format.format(id)
        else:
            url = url_format

        resp = self.session.get(url, params=parameters)
        raw = resp.json()

        # NOTE: Will we need this in future?
        if "@odata.context" in raw:
            del raw["@odata.context"]

        if "value" in raw:
            if isinstance(raw["value"], list):
                raw = raw["value"]

        return raw

    def get_groups(self, filter=None, top=None, skip=None):
        """
        Return a list of workspaces the user has access to.

        Parameters
        ----------
        `filter` : `str`
            Filters the results, based on a boolean condition, by default None
        `top` : `int`
            Returns only the first n results, by default None
        `skip` : `int`
            Skip the first n results, by default None

        Returns
        -------
        `list`
            List of `Group` objects.
        """

        resource = "https://api.powerbi.com/v1.0/myorg/groups"
        params = {
            "$filter": filter,
            "$top": top,
            "$skip": skip,
        }
        raw = self._get_resource(resource, parameters=params)

        return [Group.from_raw(raw=group) for group in raw]

    def get_datasets_in_group(self, group):
        if isinstance(group, Group):
            id = group.id
        else:
            id = group

        resource = "https://api.powerbi.com/v1.0/myorg/groups/{0}/datasets"
        raw = self._get_resource(resource, id)

        return [Dataset.from_raw(raw=dataset) for dataset in raw]

    def get_refresh_history(self, dataset, top=None):
        """
        Returns the refresh history for the specified dataset.

        Parameters
        ----------
        `dataset` : `str`
            The Dataset Id or Dataset Object to get the refresh history for.
        `top` : `int`
            The requested number of entries in the refresh history. If not provided,
            the default is the last available 500 entries.

        Returns
        -------
        `list`
            List of `Refresh` objects.

        Raises
        ------
        `TypeError`
            If the `Dataset` is not refreshable.
        """

        # TODO: What if they pass in a partially constructed Dataset obj?
        
        # If string, we need to find out if the dataset is refreshable.
        if isinstance(dataset, str):
            dataset = self.get_dataset(dataset)

        if not dataset.is_refreshable:
            raise TypeError(
                "Dataset is not refreshable. Dataset id: {dataset.id}. Dataset name: {dataset.name}."
            )

        resource = "https://api.powerbi.com/v1.0/myorg/datasets/{0}/refreshes"
        params = {
            "$top": top,
        }

        raw = self._get_resource(resource, dataset.id, parameters=params)

        return [Refresh.from_raw(raw=refresh) for refresh in raw]

    def get_dataset(self, dataset_id):
        """
        Returns the specified dataset.

        Parameters
        ----------
        `dataset_id` : `str`
            Id of the Dataset object to get.

        Returns
        -------
        `Dataset`
            Dataset matching the specified Id.
        """

        # TODO: What if the Dataset isn't found?

        resource = "https://api.powerbi.com/v1.0/myorg/datasets/{0}"
        raw = self._get_resource(resource, dataset_id)

        return Dataset.from_raw(raw=raw)

    def get_activity_events(self, start_date_time, end_date_time, filter=None):
        """
        Returns a list of Activity Events on the Power BI Instance for a
        given time period. Activity Events can include viewing reports,
        downloading reports, modifying datasets, etc.

        Note: Due to a limitation of the Power BI Rest API, the start and
        end date times must be for the same UTC day. In other words, this
        method can only query the API one day at a time.

        If the number of Activity Events exceeds the API's limit, then this
        method will continue to call the API until all records are retrieved.

        Parameters
        ----------
        `start_date_time` : `str`
            Start date and time of the time period for audit event results. 
            Must be in ISO 8601 compliant UTC format: 'yyyy-mm-ddThh:mm:ss.SSSZ'
        `end_date_time` : `str`
            End date and time of the time period for audit event results. Must 
            be in ISO 8601 compliant UTC format: 'yyyy-mm-ddThh:mm:ss.SSSZ'
        `filter` : `str`
            Filters the results based on a boolean condition, using 'Activity',
            'UserId', or both properties. Supports only 'eq' and 'and' operators.

        Returns
        -------
        `list`
            List of `ActivityEvent` objects.

        Raises
        ------
        `ValueError`
            If supplied start and end times are invalid.
        """

        # TODO: Add support for passing in datetime types
        # TODO: Implment logging. Especially if this is making continuous calls.

        resource = "https://api.powerbi.com/v1.0/myorg/admin/activityevents"
        params = {
            # API expects dtms to be enclosed in single quotes.
            "startDateTime": "'" + start_date_time + "'",
            "endDateTime": "'" + end_date_time + "'",
            "$filter": filter,
        }

        raw = self._get_resource(resource, parameters=params)

        # Handle json errors
        if "error" in raw:
            error = raw["error"]
            message = error["message"]

            if "within the same UTC day" in message:
                raise ValueError(
                    "start_date_time and end_date_time must be on the same UTC day."
                )

        # API will return continuation token when further requests needed to
        # get all data. "continuationToken" will be null when no more data.

        activity_events = raw["activityEventEntities"]
        continuation_token = raw["continuationToken"]
        continuation_uri = raw["continuationUri"]

        while continuation_token:
            raw = self._get_resource(continuation_uri)
            activity_events.extend(raw["activityEventEntities"])
            continuation_token = raw["continuationToken"]
            continuation_uri = raw["continuationUri"]

        return activity_events



__title__ = "pbipy"
__description__ = "A Power BI Rest API wrapper for Python."
__url__ = ""
__version__ = "0.1.0"
__author__ = "Andrew Villazon"
__author_email__ = "andrew.villazon@gmail.com"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2022 Andrew Villazon"


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command
from setuptools.command.test import test as TestCommand


requires = [
    "python-dateutil==2.8.2",
    "requests==2.28.1",
]

extras = {"dev":["black"]}

test_requirements = [
    "pytest==7.1.3",
    "responses==0.21.0",
]

about = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "pbipy", "__version__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), about)

with open("README.md", "r", encoding="utf-") as f:
    readme = f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    author=about["__author__"],
    author_email=about["__author_email__"],
    python_requires=">=3.10",
    url=about["__url__"],
    packages=find_packages(
        exclude=[
            "tests",
            "*.tests",
            "*.tests.*",
            "tests.*",
        ]
    ),
    install_requires=requires,
    extras_require=extras,
    include_package_data=True,
    license=about["__license__"],
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Intended Audience :: System Administrators",
    ],
    tests_require=test_requirements,
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
        "test": PyTest,
    },
)
